##
# File:  ReleaseFileNames.py
# Date:  Nov-2019  J. Berrisford
#
# Updated:
#  15-Nov-2019 E. Peisach  Code cleanups
#
##
"""
Methods for finding file names for entries within the for_release area of ftp archive

In general the methods take accession and for_release flags.  for_release indicates the for_release directory.

"""


class ReleaseFileNames:
    def __init__(self):
        #                context   public    for_rel   gzip_pub, gzip_rel

        self.__mapping = {
            "model": ["{}.cif", "{}.cif", True, True],
            "sf": ["r{}sf.ent", "{}-sf.cif", True, False],
            "cs": ["{}_cs.str", "{}_cs.str", True, False],
            "nmr_data": ["{}_nmr-data.str", "{}_nmr-data.str", True, True],
            "emdxml": ["{}-v30.xml", "{}_v3.xml", False, False],
            "emdmap": ["{}.map", "{}.map", True, True],
            "emdfsc": ["{}_fsc.xml", "{}_fsc.xml", False, False],
            "validpdf": ["{}_validation.pdf", "{}_validation.pdf", False, False],
            "validpdffull": ["{}_full_validation.pdf", "{}_full_validation.pdf", False, False],
            "validxml": ["{}_validation.xml", "{}_validation.xml", False, False],
            "validcif": ["{}_validation.cif", "{}_validation.cif", False, False],
            "validpng": ["{}_multipercentile_validation.png", "{}_multipercentile_validation.png", False, False],
            "validsvg": ["{}_multipercentile_validation.svg", "{}_multipercentile_validation.svg", False, False],
            "valid2fo": ["{}_validation_2fo-fc_map_coef.cif", "{}_validation_2fo-fc_map_coef.cif", False, False],
            "validfo": ["{}_validation_fo-fc_map_coef.cif", "{}_validation_fo-fc_map_coef.cif", False, False],
            "validimagetar": ["{}_validation_images.tar", "{}_validation_images.tar", False, False],
        }

        # public for_rel
        self.__accession_remap = {
            "emdxml": ["hyphen", "underscore"],
            "emdmap": ["underscore", "underscore"],
            "emdfsc": ["underscore", "underscore"],
        }

    @staticmethod
    def __get_emdb_number(accession):
        """gets the EMDB number from the accession"""
        return accession[4:]
        # return accession.split("-")[-1]

    def __get_emdb_underscore_format(self, accession):
        """returns lower case emdb accession with underscore"""
        return "emd_{}".format(self.__get_emdb_number(accession))

    def __get_emdb_hyphen_format(self, accession):
        """returns lower case emdb accessin with hypen"""
        return "emd-{}".format(self.__get_emdb_number(accession))

    def get_lower_emdb_hyphen_format(self, accession):
        return self.__get_emdb_hyphen_format(accession)

    def get_lower_emdb_underscore_format(self, accession):
        return self.__get_emdb_underscore_format(accession)

    def __process_remap(self, remap_type, accession):
        """looks up the accession remapping in __accession_remap"""
        if remap_type == "hyphen":
            return self.__get_emdb_hyphen_format(accession)
        elif remap_type == "underscore":
            return self.__get_emdb_underscore_format(accession)
        else:
            raise NameError("unknown EMDB file remapping: {}".format(remap_type))  # pragma: no cover

    def __do_accession_remap(self, content, accession, for_release):
        """does accession remapping"""
        if content in self.__accession_remap:
            (public, release) = self.__accession_remap[content]
            if for_release:
                accession = self.__process_remap(release, accession)
            else:
                accession = self.__process_remap(public, accession)
        return accession

    def __getfname(self, content, accession, for_release):
        """Retrieves the released content file name with compression"""
        assert content in self.__mapping
        (public, release, pub_gzip, rel_gzip) = self.__mapping[content]
        accession = self.__do_accession_remap(accession=accession, content=content, for_release=for_release)
        if for_release:
            base = release
            gzipflag = rel_gzip
        else:
            base = public
            gzipflag = pub_gzip
        suffix = ".gz" if gzipflag else ""

        fname = base.format(accession) + suffix

        return fname

    def get_model(self, accession, for_release=False):
        return self.__getfname("model", accession, for_release)

    def get_structure_factor(self, accession, for_release=False):
        return self.__getfname("sf", accession, for_release)

    def get_chemical_shifts(self, accession, for_release=False):
        return self.__getfname("cs", accession, for_release)

    def get_emdb_xml(self, accession, for_release=False):
        return self.__getfname("emdxml", accession, for_release)

    def get_emdb_map(self, accession, for_release=False):
        return self.__getfname("emdmap", accession, for_release)

    def get_emdb_fsc(self, accession, for_release=False):
        return self.__getfname("emdfsc", accession, for_release)

    def get_validation_pdf(self, accession, for_release=False):
        return self.__getfname("validpdf", accession, for_release)

    def get_validation_full_pdf(self, accession, for_release=False):
        return self.__getfname("validpdffull", accession, for_release)

    def get_validation_xml(self, accession, for_release=False):
        return self.__getfname("validxml", accession, for_release)

    def get_validation_cif(self, accession, for_release=False):
        return self.__getfname("validcif", accession, for_release)

    def get_validation_png(self, accession, for_release=False):
        return self.__getfname("validpng", accession, for_release)

    def get_validation_svg(self, accession, for_release=False):
        return self.__getfname("validsvg", accession, for_release)

    def get_validation_2fofc(self, accession, for_release=False):
        return self.__getfname("valid2fo", accession, for_release)

    def get_validation_fofc(self, accession, for_release=False):
        return self.__getfname("validfo", accession, for_release)

    def get_nmr_data(self, accession, for_release=False):
        return self.__getfname("nmr_data", accession, for_release)

    def get_validation_image_tar(self, accession, for_release=False):
        return self.__getfname("validimagetar", accession, for_release)



# pylint: disable=logging-fstring-interpolation

import os
import subprocess
import logging
import hashlib

from datetime import datetime
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional, Dict, Any


class DataMoveError(Exception):
    """Custom exception for data movement errors."""
    pass  # pylint: disable=unnecessary-pass


class DataMover(ABC):
    """Abstract base class for data movement operations."""

    def __init__(self, dry_run: bool = False):
        """
        Initialize the data mover.

        Args:
            source_path: Source directory path
            destination_path: Destination directory path
            dry_run: If True, simulate operations without actual changes
        """
        self.dry_run = dry_run

    @abstractmethod
    def sync_files(self, source_path: str, destination_path: str,
                   file_patterns: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Sync files from source to destination.

        Args:
            file_patterns: Optional list of file patterns to sync

        Returns:
            Dictionary containing sync results and statistics
        """
        pass  # pylint: disable=unnecessary-pass

    @abstractmethod
    def verify_integrity(self, source_path: str, destination_path: str, file_path: str) -> bool:
        """
        Verify file integrity after transfer.

        Args:
            file_path: Path to the file to verify

        Returns:
            True if file integrity is verified, False otherwise
        """
        pass  # pylint: disable=unnecessary-pass


class RsyncDataMover(DataMover):
    """Rsync-based data mover with integrity checks and data loss prevention."""

    def __init__(self, dry_run: bool = False, rsync_options: Optional[List[str]] = None):
        """
        Initialize the Rsync data mover.

        Args:
            dry_run: If True, simulate operations without actual changes
            rsync_options: Additional rsync options
        """
        self.dry_run = dry_run
        super().__init__(dry_run)
        self.logger = logging.getLogger(self.__class__.__name__)

        # Default rsync options for safety and integrity
        self.default_options = [
            '-avz',           # archive mode, verbose, compress
            '--update',       # only transfer files that are newer
            '--checksum',     # use checksums instead of mod-time & size
            '--partial',      # keep partially transferred files
            '--progress',     # show progress
            '--stats',        # show file transfer statistics
            '--human-readable',  # human readable numbers
            '--itemize-changes',  # itemize changes
        ]

        if dry_run:
            self.default_options.append('--dry-run')

        self.rsync_options = rsync_options or []
        self.all_options = self.default_options + self.rsync_options

    def _validate_paths(self, source_path: Path, destination_path: Path) -> None:
        """Validate source and destination paths."""
        if not source_path.exists():
            raise DataMoveError(f"Source path does not exist: {source_path}")

        if not source_path.is_dir():
            raise DataMoveError(f"Source path is not a directory: {source_path}")

        # Create destination directory if it doesn't exist
        if not self.dry_run:
            destination_path.mkdir(parents=True, exist_ok=True)

    def _calculate_checksum(self, file_path: Path) -> str:
        """Calculate MD5 checksum of a file."""
        hash_md5 = hashlib.md5()
        try:
            with open(file_path, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
            return hash_md5.hexdigest()
        except IOError as e:
            self.logger.error(f"Error calculating checksum for {file_path}: {e}")
            return ""

    def _run_rsync_command(self, source_path: Path, destination_path: Path,
                           additional_options: Optional[List[str]] = None) -> subprocess.CompletedProcess:
        """
        Execute rsync command with proper error handling.

        Args:
            source_path: Source directory path
            destination_path: Destination directory path
            additional_options: Additional rsync options for this specific run

        Returns:
            CompletedProcess object with command results
        """
        options = self.all_options.copy()
        if additional_options:
            options.extend(additional_options)

        # Ensure source path ends with / to sync contents, not the directory itself
        source_str = str(source_path).rstrip('/') + '/'
        destination_str = str(destination_path)

        cmd = ['rsync'] + options + [source_str, destination_str]

        self.logger.info(f"Executing rsync command: {' '.join(cmd)}")

        try:
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
                timeout=3600  # 1 hour timeout
            )

            if result.stdout:
                self.logger.debug(f"Rsync output: {result.stdout}")
            if result.stderr:
                self.logger.warning(f"Rsync stderr: {result.stderr}")

            return result

        except subprocess.TimeoutExpired:
            raise DataMoveError("Rsync operation timed out")  # pylint: disable=raise-missing-from
        except Exception as e:
            raise DataMoveError(f"Error executing rsync command: {e}")  # pylint: disable=raise-missing-from

    def sync_files(self, source_path: str, destination_path: str,
                   file_patterns: Optional[List[str]] = None) -> Dict[str, Any]:
        """
        Sync files using rsync with integrity checks.

        Args:
            source_path: Source directory path
            destination_path: Destination directory path
            file_patterns: Optional list of file patterns to include/exclude

        Returns:
            Dictionary containing sync results and statistics
        """
        source_path_obj = Path(source_path)
        destination_path_obj = Path(destination_path)

        self._validate_paths(source_path_obj, destination_path_obj)

        additional_options = []

        # Add include/exclude patterns if specified
        if file_patterns:
            for pattern in file_patterns:
                if pattern.startswith('!'):  # Exclude pattern
                    additional_options.extend(['--exclude', pattern[1:]])
                else:  # Include pattern
                    additional_options.extend(['--include', pattern])

        # Get pre-sync file counts for statistics
        pre_sync_stats = self._get_directory_stats(source_path_obj)

        # Execute rsync
        result = self._run_rsync_command(source_path_obj, destination_path_obj, additional_options)

        # Parse rsync exit code
        success = result.returncode == 0
        if not success:
            error_msg = f"Rsync failed with exit code {result.returncode}: {result.stderr}"
            self.logger.error(error_msg)
            raise DataMoveError(error_msg)

        # Get post-sync statistics
        post_sync_stats = self._get_directory_stats(destination_path_obj) if not self.dry_run else {}

        # Parse rsync output for transfer statistics
        transfer_stats = self._parse_rsync_stats(result.stdout)

        return {
            'success': success,
            'dry_run': self.dry_run,
            'source_path': str(source_path_obj),
            'destination_path': str(destination_path_obj),
            'pre_sync_stats': pre_sync_stats,
            'post_sync_stats': post_sync_stats,
            'transfer_stats': transfer_stats,
            'timestamp': datetime.now().isoformat()
        }

    def verify_integrity(self, source_path: str, destination_path: str, file_path: str) -> bool:
        """
        Verify file integrity by comparing checksums.

        Args:
            source_path: Source directory path
            destination_path: Destination directory path
            file_path: Relative path to the file to verify

        Returns:
            True if file integrity is verified, False otherwise
        """
        source_path_obj = Path(source_path)
        destination_path_obj = Path(destination_path)

        source_file = source_path_obj / file_path
        dest_file = destination_path_obj / file_path

        if not source_file.exists() or not dest_file.exists():
            self.logger.error(f"File missing for integrity check: {file_path}")
            return False

        # Compare file sizes first (quick check)
        if source_file.stat().st_size != dest_file.stat().st_size:
            self.logger.error(f"File size mismatch for {file_path}")
            return False

        # Compare checksums
        source_checksum = self._calculate_checksum(source_file)
        dest_checksum = self._calculate_checksum(dest_file)

        if source_checksum != dest_checksum:
            self.logger.error(f"Checksum mismatch for {file_path}")
            return False

        self.logger.info(f"Integrity verified for {file_path}")
        return True

    def get_file_info(self, file_path: str, base_path: Optional[str] = None) -> Dict[str, Any]:
        """
        Get comprehensive file information.

        Args:
            file_path: Path to the file (can be absolute or relative to base_path)
            base_path: Base path if file_path is relative

        Returns:
            Dictionary containing file information
        """
        # Handle both absolute and relative paths
        if os.path.isabs(file_path):
            full_path = Path(file_path)
        elif base_path:
            full_path = Path(base_path) / file_path
        else:
            full_path = Path(file_path)

        if not full_path.exists():
            return {'exists': False, 'path': str(full_path)}

        stat = full_path.stat()

        return {
            'exists': True,
            'path': str(full_path),
            'size': stat.st_size,
            'modified_time': datetime.fromtimestamp(stat.st_mtime).isoformat(),
            'created_time': datetime.fromtimestamp(stat.st_ctime).isoformat(),
            'checksum': self._calculate_checksum(full_path),
            'is_file': full_path.is_file(),
            'is_directory': full_path.is_dir(),
            'permissions': oct(stat.st_mode)[-3:]
        }

    def _get_directory_stats(self, path: Path) -> Dict[str, Any]:
        """Get statistics about a directory."""
        if not path.exists():
            return {'exists': False}

        file_count = 0
        dir_count = 0
        total_size = 0

        try:
            for item in path.rglob('*'):
                if item.is_file():
                    file_count += 1
                    total_size += item.stat().st_size
                elif item.is_dir():
                    dir_count += 1
        except Exception as e:
            self.logger.warning(f"Error getting directory stats: {e}")

        return {
            'exists': True,
            'path': str(path),
            'file_count': file_count,
            'directory_count': dir_count,
            'total_size_bytes': total_size,
            'total_size_human': self._human_readable_size(total_size)
        }

    def _parse_rsync_stats(self, rsync_output: str) -> Dict[str, Any]:
        """Parse rsync statistics from output."""
        stats = {}

        # Look for common rsync statistics patterns
        lines = rsync_output.split('\n')
        for line in lines:
            line = line.strip()
            if 'sent' in line and 'received' in line:
                # Parse: "sent 1,234 bytes  received 567 bytes  890.00 bytes/sec"
                stats['transfer_summary'] = line
            elif line.startswith('Number of files'):
                stats['file_summary'] = line
            elif line.startswith('Total file size'):
                stats['size_summary'] = line

        return stats

    def _human_readable_size(self, size_in_bytes: int) -> str:
        """Convert bytes to human readable format."""
        for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
            if size_in_bytes < 1024:
                return f"{size_in_bytes:.1f} {unit}"
            size_in_bytes /= 1024
        return f"{size_in_bytes:.1f} PB"



import os
import shutil
import logging
import tarfile
from fnmatch import fnmatch


logger = logging.getLogger()


class Compression:
    def __init__(self, config, dbapi) -> None:
        # injecting dbapi so the connection can be opened only once
        # by the calling code, in case of multiple entries
        self._archive_dir = os.path.join(config.get("SITE_ARCHIVE_STORAGE_PATH"), "archive")
        # maybe this should be read from a separate variable to allow this location to be separate from archive
        self._cold_archive_dir = os.path.join(config.get("SITE_ARCHIVE_STORAGE_PATH"), "cold_archive")

        if not os.path.exists(self._cold_archive_dir):
            raise Exception(f"{self._cold_archive_dir} does not exist")  # pylint: disable=broad-exception-raised)

        self._dbapi = dbapi

    def _can_be_compressed(self, dep_id: str):
        rows = self._dbapi.runSelectNQ(table="deposition", select=["notify", "locking"], where={"dep_set_id": dep_id})

        if len(rows) == 0:
            raise Exception(f"Couldn't find entry in table `status.deposition` for deposition `{dep_id}`")  # pylint: disable=broad-exception-raised)

        if rows[0][0] in ("*", "R*", "T*", "R", "TR", "T", "NT*", "NTR*", "NT", "NTR"):
            logger.warning(f"Deposition `{dep_id}` cannot be compressed due to `notify` value `{rows[0][0]}`")  # pylint: disable=logging-fstring-interpolation
            return False

        if rows[0][1].lower() == "wfm":
            logger.warning(f"Deposition `{dep_id}` cannot be compressed as it's unlocked (locking = {rows[0][1]})")  # pylint: disable=logging-fstring-interpolation
            return False

        rows = self._dbapi.runSelectNQ(table="communication", select=["status"], where={"dep_set_id": dep_id})

        if rows[0][0].lower() == "working":
            logger.warning(f"Deposition `{dep_id}` cannot be compressed as it's being processed by the WFE")  # pylint: disable=logging-fstring-interpolation
            return False

        return True

    def is_compressed(self, dep_id: str):
        dep_tarball = os.path.join(self._cold_archive_dir, f"{dep_id}.tar.gz")

        if os.path.exists(dep_tarball):
            return True

        return False

    def check_tarball(self, dep_id: str):
        dep_tarball = os.path.join(self._cold_archive_dir, f"{dep_id}.tar.gz")

        if not self.is_compressed(dep_id=dep_id):
            raise Exception(f"{dep_id} is not compressed")  # pylint: disable=broad-exception-raised)

        with tarfile.open(dep_tarball, "r:gz") as fp:
            fp.getmembers()

    def compress(self, dep_id: str, overwrite: bool = False):
        if not dep_id.startswith("D_"):
            raise Exception("Invalid deposition id")  # pylint: disable=broad-exception-raised)

        dep_archive = os.path.join(self._archive_dir, dep_id)
        dep_tarball = os.path.join(self._cold_archive_dir, f"{dep_id}.tar.gz")

        if self.is_compressed(dep_id=dep_id) and not overwrite:
            raise Exception(f"{dep_id} is already compressed. Set `overwrite` to True to overwrite it.")  # pylint: disable=broad-exception-raised)

        if not os.path.exists(dep_archive):
            raise Exception(f"Deposition {dep_id} does not exist")  # pylint: disable=broad-exception-raised)

        if not self._can_be_compressed(dep_id=dep_id):
            raise Exception(f"Deposition {dep_id} cannot be compressed")  # pylint: disable=broad-exception-raised)

        logging.info("Compressing %s to %s", dep_archive, dep_tarball)

        with tarfile.open(dep_tarball, "w:gz", debug=1) as tf:
            tf.add(dep_archive, arcname=dep_id)

        # this will throw if the file is corrupt
        self.check_tarball(dep_id=dep_id)
        shutil.rmtree(dep_archive)

        return dep_tarball

    def decompress(self, dep_id: str, overwrite: bool = False):
        dep_archive = os.path.join(self._archive_dir, dep_id)
        dep_tarball = os.path.join(self._cold_archive_dir, f"{dep_id}.tar.gz")

        if not self.is_compressed(dep_id=dep_id):
            raise Exception(f"{dep_id} is not compressed")  # pylint: disable=broad-exception-raised)

        if os.path.exists(dep_archive) and not overwrite:
            raise Exception(f"{dep_id} is already decompressed. Set `overwrite` to True to overwrite it.")  # pylint: disable=broad-exception-raised)

        with tarfile.open(dep_tarball, "r:gz") as tf:
            tf.extractall(self._archive_dir)
            logging.info(f"{dep_id} extracted successfully")  # pylint: disable=logging-fstring-interpolation

    def get_compressed_count(self):
        for _root, _dirs, files in os.walk(self._cold_archive_dir):
            tar_files = [tf for tf in files if fnmatch(tf, "*.tar.gz")]

        return len(tar_files)



"""
Utility class of generic data file methods.

File:    DataFile.py
Author:  jdw
Update:  21-Aug-2009
Version: 001
Initial version - adapted from rcsbDataFile.py  method collections.

Update:  22-Aug-2009
Version: 002  Fixed comparison function -

This software was developed as part of the Protein Structure Initiative
Structural Genomics Knowledgebase (PSI SGKB) project at Rutgers University.

Copyright (c) 2009 PSI SGKB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""

__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__version__ = "V0.002"


import sys
import os
import os.path
import time
import datetime
import filecmp
import stat
import shutil
import smtplib
import tempfile
from email.mime.text import MIMEText

##
##


class DataFile:

    """Class of utilities for managing data files."""

    def __init__(self, fPath=None, tMode=None, verbose=False):
        self.test = " "
        self.verbose = verbose
        self.vout = sys.stdout
        self.srcPathInp = None
        self.srcPath = None
        self.srcDirName = None
        self.srcFileName = None
        self.srcFileBaseName = None
        self.srcFileExt = None
        self.srcType = None
        self.srcStat = None
        self.__updateSrc(fPath)
        #
        self.dstPathInp = None
        self.dstPath = None
        self.dstDirName = None
        self.dstFileName = None
        self.dstFileBaseName = None
        self.dstFileExt = None
        self.dstType = None
        self.dstStat = None
        #
        self.__formatTimeStamp = "%Y-%m-%d:%H:%M:%S"
        self.tMode = tMode

    def __updateSrc(self, fPath):
        if fPath is not None:
            self.srcPathInp = fPath
            self.srcPath = os.path.abspath(fPath)
            (self.srcDirName, self.srcFileName) = os.path.split(self.srcPath)
            (self.srcFileBaseName, self.srcFileExt) = os.path.splitext(self.srcFileName)
            self.srcType = self.__ftype(self.srcFileExt)
            self.srcStat = None

    def __updateDst(self, fPath):
        if fPath is not None:
            self.dstPathInp = fPath
            self.dstPath = os.path.abspath(fPath)
            (self.dstDirName, self.dstFileName) = os.path.split(self.dstPath)
            (self.dstFileBaseName, self.dstFileExt) = os.path.splitext(self.dstFileName)
            self.dstType = self.__ftype(self.dstFileExt)
            self.dstStat = None

    def __ftype(self, fExt):
        if fExt == ".Z":
            ftype = "zlib"
        elif fExt == ".gz":
            ftype = "gzip"
        elif fExt == ".bz2":
            ftype = "bzip"
        else:
            ftype = None
        return ftype

    def __exists(self, fPath):
        if fPath is not None:
            if os.access(fPath, os.F_OK):
                return True
            else:
                return False
        else:
            return False

    def __stat(self, fPath):
        if fPath is not None:
            try:
                tup = os.stat(fPath)
                return tup
            except:  # noqa: E722 pylint: disable=bare-except
                return None
        else:
            return None

    def __mkdir(self, path):
        if not os.path.isdir(path):
            os.makedirs(path, 0o755)

    def __pathDivide(self, p, rest=None):
        if rest is None:
            rest = []
        (h, t) = os.path.split(p)
        if len(h) < 1:
            return [t] + rest
        if len(t) < 1:
            return [h] + rest
        return self.__pathDivide(h, [t] + rest)

    def __findCommonMembers(self, l1, l2, common=None):
        if common is None:
            common = []
        if len(l1) < 1:
            return (common, l1, l2)
        if len(l2) < 1:
            return (common, l1, l2)
        if l1[0] != l2[0]:
            return (common, l1, l2)
        return self.__findCommonMembers(l1[1:], l2[1:], common + [l1[0]])

    def __makeRelativePath(self, src, dst):
        #        self.vout.write("++INFO - src %s\n" % src)
        #        self.vout.write("++INFO - dst %s\n" % dst)
        (_common, l1, l2) = self.__findCommonMembers(self.__pathDivide(dst), self.__pathDivide(src))
        p = []
        #        if (len(l1) > 0):
        #            p = [ '../' * len(l1) ]
        if len(l1) > 1:
            p = ["../" * (len(l1) - 1)]

        p = p + l2
        return os.path.join(*p)

    def __symLinkRelative(self):
        """Internal method that makes a relative symbolic link to srcPath from dstPath.
        Intervening path components are created for the destination so this method
        is only appropriate for linking files.
        """
        if not self.srcFileExists() or not os.path.isabs(self.srcPath):
            return
        if self.dstPath is None:
            return
        if not self.dstDirExists():
            self.__mkdir(self.dstDirName)

        rPath = self.__makeRelativePath(self.srcPath, self.dstPath)

        if rPath.startswith(".."):
            # make link
            #            self.vout.write("++INFO - linking %s to %s\n" % (self.dstPath,rPath))
            if os.path.islink(self.dstPath):
                os.unlink(self.dstPath)
            return os.symlink(rPath, self.dstPath)
        else:
            self.vout.write("++ERROR - relative pathname creation failed %s\n" % rPath)
            return -1

    def __symLink(self):
        """Internal method that creates a symbolic link to srcPath named dstPath.
        Intervening path components are created for the destination so this method
        is only appropriate for linking files.
        """
        if not self.srcFileExists() or not os.path.isabs(self.srcPath):
            return
        if self.dstPath is None:
            return
        if not self.dstDirExists():
            self.__mkdir(self.dstDirName)
        if os.path.islink(self.dstPath):
            os.unlink(self.dstPath)
        return os.symlink(self.srcPath, self.dstPath)

    def __copy(self, op="copy"):
        """Internal method that copies srcPath to dstPath converting compression mode
        according to file type.
        """
        if not self.srcFileExists():
            return
        if not self.dstDirExists():
            self.__mkdir(self.dstDirName)

        if self.srcType == self.dstType and op == "copy":
            return shutil.copy2(self.srcPath, self.dstPath)
        else:
            if (self.srcType == "zlib") or (self.srcType == "gzip"):
                cmdP1 = "zcat " + self.srcPath
            elif self.srcType == "bzip":
                cmdP1 = "bzcat " + self.srcPath
            elif self.srcType is None:
                cmdP1 = "cat " + self.srcPath
            else:
                # Should provide an error
                cmdP1 = "cat " + self.srcPath

            if op == "copy":
                redir = " > "
            elif op == "append":
                redir = " >> "
            else:
                redir = " > "

            if self.dstType == "zlib":
                cmdP2 = " |  compress -cf - " + redir + self.dstPath
            elif self.dstType == "gzip":
                cmdP2 = " |  gzip " + redir + self.dstPath
            elif self.dstType == "bzip":
                cmdP2 = " | bzip2 " + redir + self.dstPath
            elif self.dstType is None:
                cmdP2 = redir + self.dstPath
            else:
                # Should log an error somehow
                cmdP2 = redir + self.dstPath

            cmd = cmdP1 + cmdP2
            return os.system(cmd)

    def __compare(self):
        """Compare srcPath to dstPath converting compression according to
           file extension (.Z, .gz, .bz ).

        Return True if the files are content equivalent and
               False otherwise.

        """
        isSame = False
        if not self.srcFileExists():
            return isSame
        if not self.dstFileExists():
            return isSame

        if self.srcType == self.dstType:
            isSame = filecmp.cmp(self.srcPath, self.dstPath, False)
        else:
            cmd = ""
            if self.srcType is None:
                f1 = self.srcPath
            elif (self.srcType == "zlib") or (self.srcType == "gzip"):
                f1 = tempfile.mktemp()
                cmd = "zcat " + self.srcPath + "> " + f1
            elif self.srcType == "bzip":
                f1 = tempfile.mktemp()
                cmd = "bzcat " + self.srcPath + "> " + f1

            if len(cmd) > 0:
                print("src command: %s" % cmd)
                os.system(cmd)

            cmd = ""
            if self.dstType is None:
                f2 = self.dstPath
            elif (self.dstType == "zlib") or (self.dstType == "gzip"):
                f2 = tempfile.mktemp()
                cmd = "zcat " + self.dstPath + "> " + f2
            elif self.dstType == "bzip":
                f2 = tempfile.mktemp()
                cmd = "bzcat " + self.dstPath + "> " + f2

            if len(cmd) > 0:
                print("dst command: %s" % cmd)
                os.system(cmd)

            if os.access(f1, os.F_OK) and os.access(f2, os.F_OK):
                isSame = filecmp.cmp(f1, f2, False)
                if f1 != self.srcPath:
                    os.remove(f1)
                if f2 != self.dstPath:
                    os.remove(f2)
            else:
                sys.stdout.write("++ERROR - comparison failed for %s with %s \n" % (f1, f2))

        return isSame

    def __remove(self):
        """Internal method that removes srcPath if it exists."""
        if not self.srcFileExists():
            return True
        if os.path.islink(self.srcPath):
            return os.unlink(self.srcPath)
        elif os.path.isfile(self.srcPath):
            return os.unlink(self.srcPath)
        elif os.path.isdir(self.srcPath):
            return shutil.rmtree(self.srcPath, True)

    def __move(self):
        """Internal method that moves srcPath to dstPath."""
        if not self.srcFileExists():
            return
        if not self.dstDirExists():
            self.__mkdir(self.dstDirName)
        if self.srcType == self.dstType:
            return shutil.move(self.srcPath, self.dstPath)

    def __timeStampCopy(self):
        """Reset the atime/mtimes of the destination using those of the
        source file.
        """
        if self.srcStat is None:
            self.srcStat = self.__stat(self.srcPath)
        if self.srcStat is not None and self.__exists(self.dstPath):
            times = (self.srcStat[stat.ST_ATIME], self.srcStat[stat.ST_MTIME])
            os.utime(self.dstPath, times)

    def __setTimeStamp(self, fPath, tObj):
        """Reset the atime/mtimes of the input file using the input timestamp."""
        if tObj is None or not self.__exists(fPath):
            return
        mtime = int(time.mktime(tObj.timetuple()))
        times = (mtime, mtime)
        os.utime(fPath, times)

    def __setFileMode(self, fPath, mode):
        """Reset the file mode of the input file using the input integer mode (e.g. 0644)"""
        try:
            os.chmod(fPath, mode)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            return False

    def setSrcFileMode(self, mode):
        if not self.srcFileExists():
            return False
        return self.__setFileMode(self.srcPath, mode)

    def setDstFileMode(self, mode):
        if not self.dstFileExists():
            return False
        return self.__setFileMode(self.dstPath, mode)

    def __setDstTimeStamp(self):
        if self.tMode is not None and self.__exists(self.dstPath):
            if self.tMode == "preserve":
                self.__timeStampCopy()
            elif self.tMode == "today":
                tObj = datetime.datetime.today()
                self.__setTimeStamp(self.dstPath, tObj)
            elif self.tMode == "yesterday":
                tObj = datetime.datetime.today() + datetime.timedelta(days=-1)
                self.__setTimeStamp(self.dstPath, tObj)
            elif self.tMode == "tomorrow":
                tObj = datetime.datetime.today() + datetime.timedelta(days=1)
                self.__setTimeStamp(self.dstPath, tObj)
            elif self.tMode == "lastweek":
                tObj = datetime.datetime.today() + datetime.timedelta(days=-7)
                self.__setTimeStamp(self.dstPath, tObj)

    #

    def src(self, fPath):
        """Set or reset the source file path"""
        self.__updateSrc(fPath)

    def dst(self, fPath):
        """Set or reset the destination file path"""
        self.__updateDst(fPath)

    def timeMode(self, tMode):
        """Sets the behavior of the timestamping of destination files: one of None, preserve, today, ..."""
        self.tMode = tMode

    def pr(self, fh=sys.stdout):
        """Output the information about source and destination files"""
        if self.srcPath is not None:
            fh.write("Source path (input):        %s\n" % self.srcPathInp)
            fh.write("Source path:                %s\n" % self.srcPath)
            fh.write("Source directory:           %s\n" % self.srcDirName)
            fh.write("Source file name:           %s\n" % self.srcFileName)
            fh.write("Source file extension:      %s\n" % self.srcFileExt)
            fh.write("Source file type :          %s\n" % str(self.srcType))
            #  lt = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
            if self.srcStat is None:
                self.srcStat = self.__stat(self.srcPath)
            if self.srcStat is not None:
                fh.write("Source creation time:       %s\n" % time.ctime(self.srcStat[stat.ST_CTIME]))
                fh.write("Source modification time:   %s\n" % time.ctime(self.srcStat[stat.ST_MTIME]))
                fh.write("Source access time:         %s\n" % time.ctime(self.srcStat[stat.ST_ATIME]))
                fh.write("Source size (bytes):        %d\n" % self.srcStat[stat.ST_SIZE])
                fh.write("Source uid/gid:             %d|%d\n" % (self.srcStat[stat.ST_UID], self.srcStat[stat.ST_GID]))
                fh.write("Source mode:                %o\n" % self.srcStat[stat.ST_MODE])

        if self.dstPath is not None:
            fh.write("Destination path (input):   %s\n" % self.dstPathInp)
            fh.write("Destination path:           %s\n" % self.dstPath)
            fh.write("Destination directory:      %s\n" % self.dstDirName)
            fh.write("Destination file name:      %s\n" % self.dstFileName)
            fh.write("Destination file extension: %s\n" % self.dstFileExt)
            fh.write("Destination file type:      %s\n" % str(self.dstType))
            if self.dstStat is None:
                self.dstStat = self.__stat(self.dstPath)
            if self.dstStat is not None:
                fh.write("Destination creation time:       %s\n" % time.ctime(self.dstStat[stat.ST_CTIME]))
                fh.write("Destination modification time:   %s\n" % time.ctime(self.dstStat[stat.ST_MTIME]))
                fh.write("Destination access time:         %s\n" % time.ctime(self.dstStat[stat.ST_ATIME]))
                fh.write("Destination size (bytes):        %d\n" % self.dstStat[stat.ST_SIZE])
                fh.write("Destination uid/gid:             %d|%d\n" % (self.dstStat[stat.ST_UID], self.dstStat[stat.ST_GID]))
                fh.write("Destination mode:                %o\n" % self.dstStat[stat.ST_MODE])

    def copy(self, dstPath=None):
        """Copies srcPath to dstPath converting compression mode
        according to file type.
        """
        self.dst(dstPath)
        self.__copy(op="copy")
        self.__setDstTimeStamp()

    def append(self, dstPath=None):
        """Appends srcPath to dstPath converting compression mode
        according to file type.
        """
        self.dst(dstPath)
        self.__copy(op="append")
        self.__setDstTimeStamp()

    def compare(self, dstPath=None):
        """Compare srcPath to dstPath converting compression mode
        according to file type.

        Returns True if the files are content equivalent or False otherwise
        """
        self.dst(dstPath)
        return self.__compare()

    def remove(self):
        """Removes srcPath."""
        self.__remove()

    def srcFileSize(self):
        if self.srcStat is None:
            self.srcStat = self.__stat(self.srcPath)
        if self.srcStat is not None:
            return self.srcStat[stat.ST_SIZE]
        else:
            return 0

    def dstFileSize(self):
        if self.dstStat is None:
            self.dstStat = self.__stat(self.dstPath)
        if self.dstStat is not None:
            return self.dstStat[stat.ST_SIZE]
        else:
            return 0

    def move(self, dstPath=None):
        """Moves (renames) srcPath to dstPath."""
        self.dst(dstPath)
        self.__move()
        self.__setDstTimeStamp()

    def symLinkRelative(self, dstPath=None):
        """Creates a relative symbolic link to srcPath at dstPath."""
        self.dst(dstPath)
        self.__symLinkRelative()

    #        self.__setDstTimeStamp()

    def symLink(self, dstPath=None):
        """Create a symbolic link to srcPath at dstPath."""
        self.dst(dstPath)
        self.__symLink()

    #        self.__setDstTimeStamp()

    def srcFileExists(self):
        """Return True if source file exists or False otherwise."""
        return self.__exists(self.srcPath)

    def dstFileExists(self):
        """Return True if destination file exists or False otherwise."""
        return self.__exists(self.dstPath)

    def dstDirExists(self):
        """Return True if destination file exists or False otherwise."""
        return self.__exists(self.dstDirName)

    def srcModTime(self):
        """Return modification time of the source file as the number of seconds
        since the epoch (ie. 1970-01-01).
        """
        if self.srcStat is None:
            self.srcStat = self.__stat(self.srcPath)
        if self.srcStat is not None:
            return self.srcStat[stat.ST_MTIME]
        else:
            return None

    def srcModTimeStamp(self):
        """Return modification time of the source file as the number of seconds
        since the epoch (ie. 1970-01-01).
        """
        if self.srcStat is None:
            self.srcStat = self.__stat(self.srcPath)
        if self.srcStat is not None:
            return time.strftime(self.__formatTimeStamp, time.localtime(self.srcStat[stat.ST_MTIME]))
        else:
            return None

    def newerThan(self, fPath):
        """Return True if srcPath has been modified more recently than fPath.
        If the comparison cannot be made 'None' is returned.
        """
        if fPath is None:
            return None
        fTup = self.__stat(fPath)
        if fTup is not None:
            modTimeF = fTup[stat.ST_MTIME]
        else:
            return None

        modTime = self.srcModTime()
        if modTime is not None:
            if modTime > modTimeF:
                return True
            else:
                return False
        else:
            # no stat for srcPath
            return None

    def eMail(self, toAddr, fromAddr, subject):  # pragma: no cover
        """Internal method to mail srcPath file as text."""
        if not self.srcFileExists():
            return

        cmd = ""
        if self.srcType is None:
            f1 = self.srcPath
        elif (self.srcType == "zlib") or (self.srcType == "gzip"):
            f1 = tempfile.mktemp()
            cmd = "zcat " + self.srcPath + "> " + f1
        elif self.srcType == "bzip":
            f1 = tempfile.mktemp()
            cmd = "bzcat " + self.srcPath + "> " + f1
        if len(cmd) > 0:
            os.system(cmd)

        # Create a text/plain message
        fp = open(f1, "rb")
        msg = MIMEText(fp.read())
        fp.close()
        msg["Subject"] = subject
        msg["From"] = fromAddr
        msg["To"] = toAddr
        s = smtplib.SMTP()
        s.connect()
        s.sendmail(fromAddr, [toAddr], msg.as_string())
        s.close()
        if f1 != self.srcPath:
            os.remove(f1)

    def setDstMTimeYYYYMMDD(self, dateYYYYMMDD):
        """Set access and modification times to the input date stamp."""

    def setDstMTime(self, refPath):
        """Set access and modification times to the input date stamp."""



"""
Pretty print a nest structure of dictionaries and lists such as
a the object collection produced by the entryFactory.

File:    FormatOut.py
Author:  jdw
Date:    24-Mar-2006
Version: 0.001
Split off from utils.py and converted to a class with string buffer.

Update:  5-Feb-2010 jdw     Add stream method -
        17-Aug-2014 jdw     Fix bool type stream formatting

"""
import sys
import traceback

MAX_INDENT = 100
SPACE = " " * MAX_INDENT


class FormatOut(object):
    """"""

    def __init__(self):
        self.__buffer = []

    def writeStream(self, fObj):
        try:
            fObj.write("".join(self.__buffer))
        except:  # noqa: E722 pylint: disable=bare-except # pragma: no cover
            traceback.print_exc(file=sys.stderr)

    def write(self, filename):
        fH = open(filename, "w")
        fH.writelines(self.__buffer)
        fH.close()

    def clear(self):
        self.__buffer = []

    def indent(self, strIn, indent=0):
        if indent > 0 and indent < MAX_INDENT:
            self.__buffer.append("%s%s" % (SPACE[1:indent], strIn))
        else:
            self.__buffer.append(strIn)

    def autoFormat(self, name, thing, indent=0, indentIncr=3):
        """Print utility for dictionaries of factory data."""
        inOt = str(type(thing)).lower()
        ind = indent + 0
        indInc = indentIncr
        if (
            inOt.find("str") > 0
            or inOt.find("int") > 0
            or inOt.find("unicode") > 0
            or inOt.find("bytes") > 0
            or inOt.find("bool") > 0
            or inOt.find("float") > 0
            or inOt.find("long") > 0
            or inOt.find("date") > 0
            or inOt.find("time") > 0
        ):
            if len(str(thing)) > 0:
                str1 = "%-20s = %s\n" % (name, str(thing))
                self.indent(str1, ind)
        elif inOt.find("list") > 0 or inOt.find("tuple") > 0:
            if len(thing) > 0:
                if name is not None and len(name) > 0:
                    self.indent("\n", ind)
                    str1 = "CONTENTS OF LIST: %s\n" % name
                    self.indent(str1, ind)
                iEl = 0
                for el in thing:
                    lab = name + " list index [" + str(iEl) + "]"
                    iEl += 1
                    self.autoFormat(lab, el, ind + indInc)
        elif inOt.find("dict") > 0:
            if len(thing) > 0:
                if name is not None and len(name) > 0:
                    self.indent("\n", ind)
                    str1 = "CONTENTS OF DICTIONARY: %s\n" % name
                    self.indent(str1, ind)
                keys = list(thing.keys())
                keys.sort()
                for k in keys:
                    self.autoFormat(str(k), thing[k], ind + indInc)
        elif inOt.find("module") > 0 or inOt.find("instance") > 0:
            if name is not None and len(name) > 0:
                self.indent("\n", ind)
                str1 = "CONTENTS OF: %s\n" % name
                self.indent(str1, ind)
            #
            # check if exists and is_callable..
            #
            thing.writeDetails(self, ind + indInc)
        elif inOt.find("nonetype") > 0:
            pass
        else:
            self.indent("\nAUTOFORMAT: CANNOT PRINT %s TYPE %s\n" % (name, inOt), ind)


def unitTest1(fileName="formatOut.log"):  # pragma: no cover
    list_in = ["L1", "L2", "L3", "L4", "L5"]
    tuple_in = ("T1", "T2", "T3", "T4", "T5")
    dict_in = {}
    for i in range(1, 10):
        t = ["D1", list_in, tuple_in]
        dict_in[i] = t

    out = FormatOut()
    out.autoFormat("unitTest1 results", dict_in, 3, 3)
    out.write(fileName)


if __name__ == "__main__":  # pragma: no cover
    unitTest1("formatOut.log")



##
# File:  ValidateXml.py
# Date:  14-Aug-2014
# Updates:
##
"""
Parse validation XML report

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) 2014 wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "Zukang Feng"
__email__ = "zfeng@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

from xml.dom import minidom
import sys
import traceback


class ValidateXml(object):
    """Class responsible for parsing validation XML report"""

    def __init__(self, FileName=None, verbose=False, log=sys.stderr):  # pylint: disable=unused-argument
        self.__xmlFile = FileName
        self.__doc = minidom.parse(self.__xmlFile)
        #
        self.clashMap = {}
        self.clashOutliers = []
        self.summaryValues = dict()
        self.__outlierMap = {}
        self.__outlierResult = {}
        self.__calculated_completeness = ""
        self.__number_of_errors_while_mapping = 0
        self.__number_of_warnings_while_mapping = 0
        self.__not_found_in_structure_cs_list = []
        self.__not_found_residue_in_structure_cs_list = []
        self.__cs_outlier_list = []
        self.__cs_referencing_offset_list = []
        self.__has_cs_referencing_offset_flag = False
        #
        self.__getOutlierDefinition()
        self.__parse()

    def getOutlier(self, Type):
        """"""
        if Type in self.__outlierResult:
            return self.__outlierResult[Type]
        #
        return []

    def getClashOutliers(self):
        """"""
        return self.clashOutliers

    def getCalculatedCompleteness(self):
        """"""
        return self.__calculated_completeness

    def getCsMappingErrorNumber(self):
        """"""
        return self.__number_of_errors_while_mapping

    def getCsMappingWarningNumber(self):
        """"""
        return self.__number_of_warnings_while_mapping

    def getNotFoundInStructureCsList(self):
        """"""
        return self.__not_found_in_structure_cs_list

    def getNotFoundResidueInStructureCsList(self):
        """"""
        return self.__not_found_residue_in_structure_cs_list

    def getCsOutliers(self):
        """"""
        return self.__cs_outlier_list

    def getCsReferencingOffsetFlag(self):
        """"""
        return self.__has_cs_referencing_offset_flag

    def getSummary(self):
        return self.summaryValues

    def __getOutlierDefinition(self):
        """"""
        self.__outlierMap["torsion-outlier"] = ["phi", "psi"]
        self.__outlierMap["mog-ring-outlier"] = ["atoms", "mean", "mindiff", "stdev", "numobs"]
        self.__outlierMap["mog-angle-outlier"] = ["atoms", "mean", "mindiff", "stdev", "numobs", "Zscore", "obsval"]
        self.__outlierMap["mog-torsion-outlier"] = ["atoms", "mean", "mindiff", "stdev", "numobs", "obsval"]
        self.__outlierMap["mog-bond-outlier"] = ["atoms", "mean", "mindiff", "stdev", "numobs", "Zscore", "obsval"]
        self.__outlierMap["chiral-outlier"] = ["atom", "problem"]
        self.__outlierMap["plane-outlier"] = ["omega", "improper", "planeRMSD", "type"]
        self.__outlierMap["bond-outlier"] = ["atom0", "atom1", "mean", "stdev", "obs", "z", "link"]
        self.__outlierMap["angle-outlier"] = ["atom0", "atom1", "atom2", "mean", "stdev", "obs", "z", "link"]
        self.__outlierMap["clash"] = ["atom", "cid", "clashmag", "dist"]

    def __parse(self):
        """"""
        self.__processGlobalValues()
        #
        self.__processChemcalShiftList()
        #
        items = ["model", "ent", "chain", "resname", "resnum", "icode"]  # , 'ligand_geometry_outlier', 'ligand_density_outlier' ]
        #
        for node in self.__doc.getElementsByTagName("ModelledSubgroup"):
            if node.nodeType != node.ELEMENT_NODE:
                continue
            #
            residueInfo = {}
            if node.hasAttribute("rama"):
                val = node.getAttribute("rama").strip()
                if val == "OUTLIER":
                    if not residueInfo:
                        residueInfo = self.__getMapInfo(node, items)
                    #
                    dirmap = self.__getMapInfo(node, self.__outlierMap["torsion-outlier"])
                    outlier = residueInfo.copy()
                    outlier.update(dirmap)
                    if "torsion-outlier" in self.__outlierResult:
                        self.__outlierResult["torsion-outlier"].append(outlier)
                    else:
                        listout = []
                        listout.append(outlier)
                        self.__outlierResult["torsion-outlier"] = listout
                    #
                #
            #
            if node.hasAttribute("rsrz"):
                val = node.getAttribute("rsrz").strip()
                if float(val) > 5:
                    if not residueInfo:
                        residueInfo = self.__getMapInfo(node, items)
                    #
                    outlier = residueInfo.copy()
                    outlier["rsrz"] = val
                    if "polymer-rsrz-outlier" in self.__outlierResult:
                        self.__outlierResult["polymer-rsrz-outlier"].append(outlier)
                    else:
                        listout = []
                        listout.append(outlier)
                        self.__outlierResult["polymer-rsrz-outlier"] = listout
                    #
                #
            #
            if node.hasAttribute("ligRSRZ"):
                val = node.getAttribute("ligRSRZ").strip()
                if float(val) > 5:
                    if not residueInfo:
                        residueInfo = self.__getMapInfo(node, items)
                    #
                    outlier = residueInfo.copy()
                    outlier["ligRSRZ"] = val
                    if "ligand-rsrz-outlier" in self.__outlierResult:
                        self.__outlierResult["ligand-rsrz-outlier"].append(outlier)
                    else:
                        listout = []
                        listout.append(outlier)
                        self.__outlierResult["ligand-rsrz-outlier"] = listout
                    #
                #
            #

            for childnode in node.childNodes:
                if childnode.nodeType != node.ELEMENT_NODE:
                    continue
                #
                if childnode.tagName not in self.__outlierMap:
                    continue
                #
                if not residueInfo:
                    residueInfo = self.__getMapInfo(node, items)
                #
                dirmap = self.__getMapInfo(childnode, self.__outlierMap[childnode.tagName])
                # jmb - removed cut off in reporting outliers in standard bond lengths and bond angles. Uses validation XML cut off instead.
                # if childnode.tagName == 'bond-outlier' or childnode.tagName == 'angle-outlier':
                #    if abs(float(dirmap['z'])) <= 10:
                #        continue
                #
                # skip bonds between standard residues in reporting,
                # these are better captured in pdbx_validate_polymer_linkage
                # - the validation is limited to a maxiumum distance
                # of 1.999 Angstroms in the validation report for unusual bonds
                if childnode.tagName == "bond-outlier":
                    if dirmap["atom0"] in ("C", "O3'") and dirmap["atom1"] in ("N", "P"):
                        continue

                if childnode.tagName == "mog-angle-outlier" or childnode.tagName == "mog-bond-outlier":
                    if abs(float(dirmap["Zscore"])) <= 10:
                        continue
                    #
                #
                if childnode.tagName == "clash":
                    atom_dict = {
                        "dist": dirmap["dist"],
                        "atom": dirmap["atom"],
                        "clashmag": dirmap["clashmag"],
                        "chain": node.getAttribute("chain").strip(),
                        "model": node.getAttribute("model").strip(),
                        "altcode": node.getAttribute("altcode").strip(),
                        "resnum": node.getAttribute("resnum").strip(),
                        "resname": node.getAttribute("resname").strip(),
                    }
                    if float(dirmap["dist"]) < 2.2:
                        self.clashMap.setdefault(dirmap["cid"], []).append(atom_dict)

                outlier = residueInfo.copy()
                outlier.update(dirmap)
                if childnode.tagName in self.__outlierResult:
                    self.__outlierResult[childnode.tagName].append(outlier)
                else:
                    listout = []
                    listout.append(outlier)
                    self.__outlierResult[childnode.tagName] = listout
                #
            #
        #

        for node in self.__doc.getElementsByTagName("ModelledEntityInstance"):
            if node.nodeType != node.ELEMENT_NODE:
                continue

            if node.hasAttribute("average_residue_inclusion"):
                try:
                    if float(node.getAttribute("average_residue_inclusion").strip()) < 0.1:
                        if "chain_average_residue_inclusion" not in self.__outlierResult:
                            self.__outlierResult["chain_average_residue_inclusion"] = []
                        self.__outlierResult["chain_average_residue_inclusion"].append(
                            {
                                "chain": node.getAttribute("chain").strip(),
                                "model": node.getAttribute("model").strip(),
                                "average_residue_inclusion": float(node.getAttribute("average_residue_inclusion").strip()) * 100,
                            }
                        )
                except ValueError:
                    pass

        if self.clashMap:
            for o in self.clashMap:
                # check that there are two clashes
                if len(self.clashMap[o]) > 1:
                    clashitem = {
                        "res1model": self.clashMap[o][0]["model"],
                        "res1num": self.clashMap[o][0]["resnum"],
                        "res1name": self.clashMap[o][0]["resname"],
                        "res1chain": self.clashMap[o][0]["chain"],
                        "res1alt": self.clashMap[o][0]["altcode"],
                        "res1atom": self.clashMap[o][0]["atom"],
                        "res2model": self.clashMap[o][1]["model"],
                        "res2num": self.clashMap[o][1]["resnum"],
                        "res2chain": self.clashMap[o][1]["chain"],
                        "res2alt": self.clashMap[o][1]["altcode"],
                        "res2name": self.clashMap[o][1]["resname"],
                        "res2atom": self.clashMap[o][1]["atom"],
                        "dist": self.clashMap[o][0]["dist"],
                        "clashmag": self.clashMap[o][0]["clashmag"],
                    }
                    #
                    if clashitem["res1atom"][0] != "H" and clashitem["res2atom"][0] != "H":
                        self.clashOutliers.append(clashitem)
                    #
                #
            #
        #

    def __getSummaryValues(self):
        summaryList = [
            "DCC_Rfree",
            "clashscore",
            "percent-RSRZ-outliers",
            "percent-rama-outliers",
            "atom_inclusion_all_atoms",
            "atom_inclusion_backbone",
            "contour_level_primary_map",
        ]

        Entry = self.__doc.getElementsByTagName("Entry")[0]
        for item in summaryList:
            if Entry.getAttribute(item) and Entry.getAttribute(item) != "NotAvailable":
                try:
                    self.summaryValues[item] = float(Entry.getAttribute(item))
                except:  # noqa: E722 pylint: disable=bare-except
                    pass

    def __processGlobalValues(self):
        """"""
        self.__getSummaryValues()
        global_values = {}
        Entry = self.__doc.getElementsByTagName("Entry")[0]
        for item in ("DCC_Rfree", "PDB-Rfree", "DCC_R", "PDB-R"):
            if Entry.getAttribute(item) and Entry.getAttribute(item) != "NotAvailable":
                global_values[item] = Entry.getAttribute(item)
            #
        for item in ("atom_inclusion_all_atoms", "atom_inclusion_backbone"):
            if Entry.getAttribute(item) and Entry.getAttribute(item) != "NotAvailable":
                value = Entry.getAttribute(item)
                if float(value) < 0.4:
                    self.__outlierResult[item] = value
        #
        if Entry.getAttribute("DataCompleteness") and Entry.getAttribute("DataCompleteness") != "NotAvailable":
            self.__calculated_completeness = Entry.getAttribute("DataCompleteness")
        #
        for global_list in (("r_free_diff", "DCC_Rfree", "PDB-Rfree"), ("r_work_diff", "DCC_R", "PDB-R")):
            if (global_list[1] in global_values) and (global_list[2] in global_values):
                if abs(float(global_values[global_list[1]]) - float(global_values[global_list[2]])) > 0.05:
                    r_map = {}
                    r_map["diff"] = str(abs(float(global_values[global_list[1]]) - float(global_values[global_list[2]])))
                    r_map[global_list[1]] = global_values[global_list[1]]
                    r_map[global_list[2]] = global_values[global_list[2]]
                    self.__outlierResult[global_list[0]] = [r_map]
                #
            #
        #

    def __processChemcalShiftList(self):
        """ chemical_shift_list.attributes = ( 'block_name', 'file_id', 'file_name', 'list_id', 'number_of_errors_while_mapping', 'number_of_mapped_shifts' \
                        'number_of_parsed_shifts', 'number_of_unparsed_shifts', 'number_of_warnings_while_mapping', 'total_number_of_shifts' )

            unmapped_chemical_shift.attributes = ( 'ambiguity', 'atom', 'chain', 'diagnostic', 'error', 'rescode', 'resnum', 'value' )

            chemical_shift_outlier.attributes = ( 'atom', 'chain', 'method', 'prediction', 'rescode', 'resnum', 'value', 'zscore' )

            referencing_offset.attributes = ( 'atom', 'number_of_measurements', 'precision', 'uncertainty', 'value' )
        """
        csList = self.__doc.getElementsByTagName("chemical_shift_list")
        if len(csList) < 1:
            return
        #
        for csNode in csList:
            if csNode.getAttribute("number_of_errors_while_mapping"):
                self.__number_of_errors_while_mapping += int(csNode.getAttribute("number_of_errors_while_mapping"))
            #
            if csNode.getAttribute("number_of_warnings_while_mapping"):
                self.__number_of_warnings_while_mapping += int(csNode.getAttribute("number_of_warnings_while_mapping"))
            #
            unmappedCsList = csNode.getElementsByTagName("unmapped_chemical_shift")
            if len(unmappedCsList) > 0:
                for unmappedNode in unmappedCsList:
                    notMappedCsResidueFlag = False
                    if unmappedNode.getAttribute("diagnostic").startswith("Residue not found in structure."):
                        notMappedCsResidueFlag = True
                    #
                    csData = []
                    for attribute in ("chain", "resnum", "rescode", "atom", "value", "error", "ambiguity"):
                        if unmappedNode.getAttribute(attribute):
                            csData.append(unmappedNode.getAttribute(attribute))
                        else:
                            csData.append("")
                        #
                    #
                    self.__not_found_in_structure_cs_list.append(csData)
                    if notMappedCsResidueFlag:
                        self.__not_found_residue_in_structure_cs_list.append(csData)
                    #
                #
            #
            csOutlierList = csNode.getElementsByTagName("chemical_shift_outlier")
            if len(csOutlierList) > 0:
                for outlierNode in csOutlierList:
                    csData = []
                    for attribute in ("chain", "resnum", "rescode", "atom", "value", "prediction", "zscore"):
                        if outlierNode.getAttribute(attribute):
                            csData.append(outlierNode.getAttribute(attribute))
                        else:
                            csData.append("")
                        #
                    #
                    self.__cs_outlier_list.append(csData)
                #
            #
            offsetList = csNode.getElementsByTagName("referencing_offset")
            if len(offsetList) > 0:
                for offsetNode in offsetList:
                    csData = []
                    for attribute in ("atom", "number_of_measurements", "precision", "uncertainty", "value"):
                        if offsetNode.getAttribute(attribute):
                            csData.append(offsetNode.getAttribute(attribute))
                        else:
                            csData.append("")
                        #
                    #
                    if csData[2] and csData[3] and csData[4]:
                        # precision = float(csData[2])
                        uncertainty = float(csData[3])
                        value = float(csData[4])
                        # if (value - precision) < uncertainty:
                        # for D_8000210797
                        if abs(value) < uncertainty:
                            continue
                        #
                        self.__has_cs_referencing_offset_flag = True
                    #
                    self.__cs_referencing_offset_list.append(csData)
                #
            #
        #

    def __getMapInfo(self, node, items):
        """"""
        mapping = {}
        for item in items:
            val = ""
            if node.hasAttribute(item):
                val = node.getAttribute(item).strip()
            #
            mapping[item] = val
        #
        return mapping


if __name__ == "__main__":  # pragma: no cover
    try:
        obj = ValidateXml(FileName=sys.argv[1])
        summary = obj.getSummary()
        print(summary)
        #       list = obj.getOutlier('polymer-rsrz-outlier')
        #       print(list)
        list1 = obj.getOutlier("r_free_diff")
        print(list1)
        ret = obj.getOutlier("atom_inclusion_all_atoms")
        print(ret)
        #       print(obj.getClashOutliers())
        print(obj.getCsMappingErrorNumber())
        print(obj.getCsMappingWarningNumber())
        print(len(obj.getNotFoundInStructureCsList()))
        print(len(obj.getCsOutliers()))
        print(obj.getCsReferencingOffsetFlag())
    except:  # noqa: E722 pylint: disable=bare-except
        traceback.print_exc(file=sys.stderr)
        sys.exit(1)



##
# File:  DataMaintenance.py
# Date:  13-Jun-2015
#
# Updates:
#    15-Jun-2015  jdw added some recover features
#    24-Oct-2016  jdw add test mode
#
##
"""
Collection of data maintenance utilities to support
purging data files for released entries.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.09"

import sys
import os
import shutil
import traceback
import glob
from datetime import datetime

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.io.locator.PathInfo import PathInfo


class DataMaintenance(object):

    """Collection of data maintenance utilities supporting
    purge and recovery of data files post release.

    This class duplicates some methods from class DataExchange for
    anticipated future use.

    """

    def __init__(self, siteId=None, testMode=False, verbose=False, log=sys.stderr):
        self.__verbose = verbose
        self.__lfh = log
        self.__siteId = siteId
        # In test mode no deletions are performed -
        self.__testMode = testMode
        self.__debug = False
        self.__sessionPath = None
        #
        self.__setup(siteId=siteId)

    def __setup(self, siteId=None):
        self.__siteId = siteId
        self.__cI = ConfigInfo(self.__siteId)
        self.__sessionPath = None
        self.__pI = PathInfo(siteId=self.__siteId, sessionPath=self.__sessionPath, verbose=self.__verbose, log=self.__lfh)

    def setSessionPath(self, inputSessionPath=None):
        """Override the path to files with fileSource="session" """
        self.__sessionPath = inputSessionPath

    def purgeLogs(self, dataSetId):
        archivePath = self.__cI.get("SITE_ARCHIVE_STORAGE_PATH")
        dirPath = os.path.join(archivePath, "archive", dataSetId, "log")
        if self.__verbose:
            self.__lfh.write("+DataMaintenance.purgeLogs() - purging logs in directory  %s\n" % (dirPath))

        if os.access(dirPath, os.W_OK):
            fpattern = os.path.join(dirPath, "*log")
            if self.__verbose:
                self.__lfh.write("+DataMaintenance.purgeLogs() - purging pattern is %s\n" % (fpattern))

            pthList = glob.glob(fpattern)
            if self.__verbose:
                self.__lfh.write("+DataMaintenance.purgeLogs() candidate path length is %d\n" % len(pthList))
            #
            for pth in pthList:
                try:
                    if self.__testMode:
                        self.__lfh.write("+DataMaintenance.purgeLogs() TEST MODE skip remove %s\n" % pth)
                    else:
                        os.remove(pth)
                except:  # noqa: E722 pylint: disable=bare-except
                    pass
            #
        return pthList

    def reversePurge(self, dataSetId, contentType, formatType="pdbx", partitionNumber=1):
        fn = self.__getArchiveFileName(dataSetId, contentType=contentType, formatType=formatType, version="none", partitionNumber=partitionNumber)

        archivePath = self.__cI.get("SITE_ARCHIVE_STORAGE_PATH")
        dirPath = os.path.join(archivePath, "archive", dataSetId)
        if self.__verbose:
            self.__lfh.write("+DataMaintenance.__setup() - purging in directory  %s\n" % (dirPath))

        if len(dirPath) < 2:
            return []
        fpattern = os.path.join(dirPath, fn + ".V*")
        if self.__verbose:
            self.__lfh.write("+DataMaintenance.__setup() - purging pattern is %s\n" % (fpattern))

        pthList = glob.glob(fpattern)
        if self.__verbose:
            self.__lfh.write("+DataMaintenance.__reversePurge() candidate length is %d\n" % len(pthList))
        #
        fList = []
        for pth in pthList:
            if not pth.endswith(".V1"):
                fList.append(pth)

        for pth in fList:
            try:
                if self.__testMode:
                    self.__lfh.write("+DataMaintenance.reversePurge() TEST MODE skip remove %s\n" % pth)
                else:
                    os.remove(pth)
            except:  # noqa: E722 pylint: disable=bare-except
                pass
            #
        return fList

    def removeWorkflowDir(self, dataSetId):
        if (dataSetId is not None) and dataSetId.startswith("D_") and (len(dataSetId) > 10):
            workflowPath = self.__cI.get("SITE_ARCHIVE_STORAGE_PATH")
            dirPath = os.path.join(workflowPath, "workflow", dataSetId)
            if os.access(dirPath, os.W_OK):
                if self.__testMode:
                    self.__lfh.write("+DataMaintenance.removeWorkflowDir() TEST MODE skip remove %s\n" % dirPath)
                else:
                    shutil.rmtree(dirPath)
                return True
            else:
                return False
        else:
            return False

    def getLogFiles(self, dataSetId, fileSource="archive"):
        pL = []
        if fileSource in ["archive"]:
            dirPath = self.__pI.getArchivePath(dataSetId)
        elif fileSource in ["deposit"]:
            dirPath = self.__pI.getDepositPath(dataSetId)
        else:
            return pL
        fpattern = os.path.join(dirPath, "*.log")
        pthList = glob.glob(fpattern)
        return pthList

    def getPurgeCandidates(self, dataSetId, wfInstanceId=None, fileSource="archive", contentType="model", formatType="pdbx", partitionNumber="1", mileStone=None, purgeType="exp"):
        """Return the latest version, and candidates for removal and compression.

        purgeType = 'exp'    use strategy for experimental and model fileSource V<last>, V2, V1
                    'other'  use strategy for other file types -- V<last> & V1

        """
        latestV = None
        rmL = []
        gzL = []
        vtL = self.getVersionFileList(
            dataSetId, wfInstanceId=wfInstanceId, fileSource=fileSource, contentType=contentType, formatType=formatType, partitionNumber=partitionNumber, mileStone=mileStone
        )
        n = len(vtL)
        if n > 0:
            latestV = vtL[0][0]
        if purgeType in ["exp"]:
            if n < 2:
                return latestV, rmL, gzL
            elif n == 2:
                gzL.append(vtL[1][0])
            elif n == 3:
                gzL.append(vtL[1][0])
                gzL.append(vtL[2][0])
            elif n > 3:
                gzL.append(vtL[n - 2][0])
                gzL.append(vtL[n - 1][0])
                for i in range(1, n - 2):
                    rmL.append(vtL[i][0])
            else:
                pass
        elif purgeType in ["report", "other"]:
            if n < 2:
                return latestV, rmL, gzL
            elif n == 2:
                gzL.append(vtL[1][0])
            elif n > 2:
                gzL.append(vtL[n - 1][0])
                for i in range(1, n - 1):
                    rmL.append(vtL[i][0])
            else:
                pass

        return latestV, rmL, gzL

    def getVersionFileListSnapshot(self, basePath, dataSetId, wfInstanceId=None, fileSource="archive", contentType="model", formatType="pdbx", partitionNumber="1", mileStone=None):
        """
        For the input content object return a list of file versions in a snapshot directory (recovery mode).

        Return:
              List of [(file path, modification date string,size),...]

        """
        pairL = []
        # basePath = '/net/wwpdb_da_data_archive/.snapshot/nightly.1/data'
        try:
            if fileSource == "archive":
                pth = self.__pI.getArchivePath(dataSetId)
                snPth = os.path.join(basePath, "archive", dataSetId)
            elif fileSource == "deposit":
                pth = self.__pI.getDepositPath(dataSetId)
                snPth = os.path.join(basePath, "deposit", dataSetId)
            else:
                pth = "."
                snPth = "."

            fPattern = self.__pI.getFilePathVersionTemplate(
                dataSetId=dataSetId,
                wfInstanceId=wfInstanceId,
                contentType=contentType,
                formatType=formatType,
                fileSource=fileSource,
                partNumber=partitionNumber,
                mileStone=mileStone,
            )
            _dir, fn = os.path.split(fPattern)
            altPattern = os.path.join(snPth, fn)
            srcL = self.__getFileListWithVersion([altPattern], sortFlag=True)
            for src in srcL:
                _d, f = os.path.split(src[0])
                dst = os.path.join(pth, f)
                if not os.access(dst, os.F_OK):
                    pairL.append((src[0], dst))

            return pairL

        except Exception as e:
            if self.__verbose:
                self.__lfh.write(
                    "+DataMaintenance.getVersionFileList() failing for data set %s instance %s file source %s err %s\n" % (dataSetId, wfInstanceId, fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)
            return []

    ##

    def getVersionFileList(self, dataSetId, wfInstanceId=None, fileSource="archive", contentType="model", formatType="pdbx", partitionNumber="1", mileStone=None):
        """
        For the input content object return a list of file versions sorted by modification time.

        Return:
              List of [(file path, modification date string,size),...]

        """
        try:
            if fileSource == "session" and self.__sessionPath is not None:
                self.__pI.setSessionPath(self.__sessionPath)

            fPattern = self.__pI.getFilePathVersionTemplate(
                dataSetId=dataSetId,
                wfInstanceId=wfInstanceId,
                contentType=contentType,
                formatType=formatType,
                fileSource=fileSource,
                partNumber=partitionNumber,
                mileStone=mileStone,
            )
            return self.__getFileListWithVersion([fPattern], sortFlag=True)
        except Exception as e:
            if self.__verbose:
                self.__lfh.write(
                    "+DataMaintenance.getVersionFileList() failing for data set %s instance %s file source %s err %r\n" % (dataSetId, wfInstanceId, fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)
            return []

    def getContentTypeFileList(self, dataSetId, wfInstanceId, fileSource="archive", contentTypeList=None):
        """
        For the input content object return a list of file versions sorted by modification time.

        Return:
              List of [(file path, modification date string,size),...]

        """
        if contentTypeList is None:
            contentTypeList = ["model"]
        try:
            if fileSource == "session" and self.__sessionPath is not None:
                self.__pI.setSessionPath(self.__sessionPath)
            fPatternList = []
            for contentType in contentTypeList:
                fPattern = self.__pI.getFilePathContentTypeTemplate(dataSetId=dataSetId, wfInstanceId=wfInstanceId, contentType=contentType, fileSource=fileSource)

                fPatternList.append(fPattern)
            if self.__debug:
                self.__lfh.write("+DataMaintenance.getContentTypeFileList() patterns %r\n" % fPatternList)
            return self.__getFileListWithVersion(fPatternList, sortFlag=True)
        except Exception as e:
            if self.__verbose:
                self.__lfh.write(
                    "+DataMaintenance.getVersionFileList() failing for data set %s instance %s file source %s error %r\n" % (dataSetId, wfInstanceId, fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)
            return []

    def getMiscFileList(self, fPatternList=None, sortFlag=True):
        if fPatternList is None:
            fPatternList = ["*"]
        return self.__getFileList(fPatternList=fPatternList, sortFlag=sortFlag)

    def getLogFileList(self, entryId, fileSource="archive"):
        if fileSource in ["archive", "wf-archive"]:
            pth = self.__pI.getArchivePath(entryId)
            fpat1 = os.path.join(pth, "*log")
            fpat2 = os.path.join(pth, "log", "*")
            patList = [fpat1, fpat2]
        elif fileSource in ["deposit"]:
            pth = self.__pI.getDepositPath(entryId)
            fpat1 = os.path.join(pth, "*log")
            fpat2 = os.path.join(pth, "log", "*")
            patList = [fpat1, fpat2]
        else:
            return []
        return self.__getFileList(fPatternList=patList, sortFlag=True)

    def __getFileListWithVersion(self, fPatternList=None, sortFlag=False):
        """
        For the input glob compatible file pattern produce a file list sorted by modification date.

        If sortFlag is set then file list is sorted by modification date (e.g. recently changes first)

        Return:
              List of [(file path, modification date string, KBytes),...]

        """
        if fPatternList is None:
            fPatternList = ["*"]
        try:
            files = []
            for fPattern in fPatternList:
                if fPattern is not None and len(fPattern) > 0:
                    files.extend(filter(os.path.isfile, glob.glob(fPattern)))

            file_ver_tuple_list = []
            for f in files:
                tL = f.split(".")
                vId = tL[-1]
                if vId.startswith("V"):
                    if vId[-1] not in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                        file_ver_tuple_list.append((f, int(vId[1:-1])))
                    else:
                        file_ver_tuple_list.append((f, int(vId[1:])))

            # Sort the tuple list by version id
            #
            if sortFlag:
                file_ver_tuple_list.sort(key=lambda x: x[1], reverse=True)

            return file_ver_tuple_list
        except Exception as e:
            if self.__verbose:
                self.__lfh.write("+DataMaintenance.getVersionFileList() failing for pattern %r error %r\n" % (fPatternList, str(e)))
                traceback.print_exc(file=self.__lfh)
            return []

    def __getFileList(self, fPatternList=None, sortFlag=True):
        """
        For the input glob compatible file pattern produce a file list sorted by modification date.

        If sortFlag is set then file list is sorted by modification date (e.g. recently changes first)

        Return:
              List of [(file path, modification date string, KBytes),...]

        """
        if fPatternList is None:
            fPatternList = ["*"]
        try:
            files = []
            for fPattern in fPatternList:
                if fPattern is not None and len(fPattern) > 0:
                    files.extend(filter(os.path.isfile, glob.glob(fPattern)))

            file_date_tuple_list = []
            for x in files:
                d = os.path.getmtime(x)
                s = float(os.path.getsize(x)) / 1000.0
                file_date_tuple = (x, d, s)
                file_date_tuple_list.append(file_date_tuple)

            # Sort the tuple list by the modification time (recent changes first)
            if sortFlag:
                file_date_tuple_list.sort(key=lambda x: x[1], reverse=True)
            rTup = []
            for fP, mT, sZ in file_date_tuple_list:
                tS = datetime.fromtimestamp(mT).strftime("%Y-%b-%d %H:%M:%S")
                rTup.append((fP, tS, sZ))
            return rTup
        except Exception as e:
            if self.__verbose:
                self.__lfh.write("+DataMaintenance.getVersionFileList() failing for patter %r error %r\n" % (fPatternList, str(e)))
                traceback.print_exc(file=self.__lfh)
            return []

    ##
    def __getArchiveFileName(self, dataSetId, wfInstanceId=None, contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
        (_fp, _d, f) = self.__targetFilePath(
            dataSetId=dataSetId,
            wfInstanceId=wfInstanceId,
            fileSource="archive",
            contentType=contentType,
            formatType=formatType,
            version=version,
            partitionNumber=partitionNumber,
            mileStone=mileStone,
        )
        return f

    # def __getInstanceFileName(self, dataSetId, wfInstanceId=None, contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
    #     (_fp, _d, f) = self.__targetFilePath(
    #         dataSetId=dataSetId,
    #         wfInstanceId=wfInstanceId,
    #         fileSource="wf-instance",
    #         contentType=contentType,
    #         formatType=formatType,
    #         version=version,
    #         partitionNumber=partitionNumber,
    #         mileStone=mileStone,
    #     )
    #     return f

    # def __getFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
    #     (fp, _d, _f) = self.__targetFilePath(
    #         dataSetId=dataSetId,
    #         wfInstanceId=wfInstanceId,
    #         fileSource=fileSource,
    #         contentType=contentType,
    #         formatType=formatType,
    #         version=version,
    #         partitionNumber=partitionNumber,
    #         mileStone=mileStone,
    #     )
    #     return fp

    def __targetFilePath(self, dataSetId, wfInstanceId=None, fileSource="archive", contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
        """Return the file path, directory path, and filen ame  for the input content object if this object is valid.

        If the file path cannot be verified return None for all values
        """
        try:
            if fileSource == "session" and self.__sessionPath is not None:
                self.__pI.setSessionPath(self.__sessionPath)
            fP = self.__pI.getFilePath(
                dataSetId=dataSetId,
                wfInstanceId=wfInstanceId,
                contentType=contentType,
                formatType=formatType,
                fileSource=fileSource,
                versionId=version,
                partNumber=partitionNumber,
                mileStone=mileStone,
            )
            dN, fN = os.path.split(fP)
            return fP, dN, fN
        except Exception as e:
            if self.__debug:
                self.__lfh.write(
                    "+DataMaintenance.__targetFilePath() failing for data set %s instance %s file source %s error %r\n" % (dataSetId, wfInstanceId, fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)

            return (None, None, None)

    # def __copyGzip(self, inpFilePath, outFilePath):
    #     """"""
    #     try:
    #         cmd = " gzip -cd  %s > %s " % (inpFilePath, outFilePath)
    #         os.system(cmd)
    #         return True
    #     except:  # noqa: E722 pylint: disable=bare-except
    #         if self.__verbose:
    #             traceback.print_exc(file=self.__lfh)
    #         return False



"""
Some functions to allow sending of email messages using the configuration
"""

import smtplib
from email.mime.text import MIMEText
import sys

from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCommunication
from wwpdb.utils.config.ConfigInfo import getSiteId


class SendEmail:
    def __init__(self, siteId=None, verbose=False, log=sys.stderr):
        self.__verbose = verbose
        self.__lfh = log
        self.__siteId = siteId if siteId is not None else getSiteId(defaultSiteId=siteId)

    def send_email(self, body, subject, to_email, from_email, relayhost="localhost"):
        """
        function to send email
        :param body: the body of the message
        :param subject: the subject of the message
        :param to_email: the to email address
        :param from_email: the from email address
        :param relayhost: The relayhost (defaults to localhost)
        :return: True if no errors, False if errors
        """
        try:
            msg = MIMEText(body)
            msg['Subject'] = subject
            msg['From'] = from_email
            msg['To'] = to_email

            # Send the message via our own SMTP server, but don't include the envelope header.
            s = smtplib.SMTP(relayhost)
            send_errors = s.sendmail(from_email, [to_email], msg.as_string())  # return is dictionary of errors
            s.quit()

            if not send_errors:
                return True
        except Exception as e:
            self.__lfh.write("Failure to send message %s\n" % e)

        return False

    def send_system_error(self, body, subject):
        """Handles the sending of a system email message based on site-config
        Returns True on succss
        """

        ciac = ConfigInfoAppCommunication(self.__siteId, self.__verbose, self.__lfh)
        reply = ciac.get_noreply_address()
        relay = ciac.get_mailserver_name()
        dest = ciac.get_system_notification_address()

        status = self.send_email(body, subject, dest, reply, relay)
        if not status and self.__verbose:
            self.__lfh.write("Error: Failure to send system error\n")

        return status



import logging
import os
import os.path

from wwpdb.utils.config.ConfigInfo import ConfigInfo

from wwpdb.io.locator.ReleaseFileNames import ReleaseFileNames

logger = logging.getLogger(__name__)


class LocalFTPPathInfo(object):
    def __init__(self, siteId=None):
        self.__siteId = siteId
        self.__cI = ConfigInfo(siteId=self.__siteId)

        self.ftp_pdb_root = self.__cI.get("SITE_PDB_FTP_ROOT_DIR")
        self.ftp_emdb_root = self.__cI.get("SITE_EMDB_FTP_ROOT_DIR")
        self.__mapping = {
            "model": "mmCIF",
            "structure_factors": "structure_factors",
            "chemical_shifts": "nmr_chemical_shifts",
            "nmr_data": "nmr_data",
        }

    def __get_mapping(self, file_type):
        return self.__mapping.get(file_type)

    def set_ftp_pdb_root(self, ftp_pdb_root):
        if ftp_pdb_root:
            self.ftp_pdb_root = ftp_pdb_root

    def set_ftp_emdb_root(self, ftp_emdb_root):
        if ftp_emdb_root is not None:
            self.ftp_emdb_root = ftp_emdb_root

    def get_ftp_pdb(self):
        if self.ftp_pdb_root:
            return os.path.join(self.ftp_pdb_root, "pdb", "data", "structures", "all")
        return ""

    def get_ftp_emdb(self):
        if self.ftp_emdb_root:
            return os.path.join(self.ftp_emdb_root, "emdb", "structures")
        return ""

    def get_model_path(self):
        return os.path.join(self.get_ftp_pdb(), self.__get_mapping("model"))

    def get_sf_path(self):
        return os.path.join(self.get_ftp_pdb(), self.__get_mapping("structure_factors"))

    def get_cs_path(self):
        return os.path.join(self.get_ftp_pdb(), self.__get_mapping("chemical_shifts"))

    def get_nmr_data_path(self):
        return os.path.join(self.get_ftp_pdb(), self.__get_mapping("nmr_data"))

    def get_model_fname(self, accession):
        model_file_name = ReleaseFileNames().get_model(accession=accession, for_release=False)
        return os.path.join(self.get_model_path(), model_file_name)

    def get_structure_factors_fname(self, accession):
        sf_file_name = ReleaseFileNames().get_structure_factor(accession=accession, for_release=False)
        return os.path.join(self.get_sf_path(), sf_file_name)

    def get_chemical_shifts_fname(self, accession):
        cs_file_name = ReleaseFileNames().get_chemical_shifts(accession=accession, for_release=False)
        return os.path.join(self.get_cs_path(), cs_file_name)

    def get_nmr_data_fname(self, accession):
        nmr_data_file_name = ReleaseFileNames().get_nmr_data(accession=accession, for_release=False)
        return os.path.join(self.get_nmr_data_path(), nmr_data_file_name)



##
# File:  ReleasePathInfo.py
# Date:  11-Nov-2019  E. Peisach
#
# Updated:
##
"""
Common methods for finding path information for release directories used in the system.

"""
__docformat__ = "restructuredtext en"
__author__ = "Ezra Peisach"
__email__ = "peisach@rcsb.rutgers.edu"
__license__ = "Apache 2.0"
__version__ = "V0.07"

import os
import os.path
import logging

from wwpdb.utils.config.ConfigInfoApp import ConfigInfoAppCommon

logger = logging.getLogger(__name__)


class ReleasePathInfo(object):
    def __init__(self, siteId=None):
        self.__siteId = siteId
        self.__cICommon = ConfigInfoAppCommon(self.__siteId)
        self.current_folder_name = "current"
        self.previous_folder_name = "previous"

    def getForReleasePath(self, subdir=None, version="current", accession=None, em_sub_path=None):
        """Returns path to for-release directory.

        Input Parameters:
        subdir: If specified,  must be one of "added", "modified", "obsolete",
                "emd", "val-reports".
        version: "current" or "previous"

        returns path, or raises exception
        """
        basedir = self.get_for_release_path()

        if version not in [self.current_folder_name, self.previous_folder_name]:
            raise NameError("version %s not allowed" % version)

        if version == self.previous_folder_name:
            basedir = os.path.join(basedir, self.previous_folder_name)

        if subdir:
            if subdir not in ["added", "modified", "obsolete", "emd", "val_reports", "em_val_reports", "val_images"]:
                raise NameError("subdir %s not allowed" % subdir)

            basedir = os.path.join(basedir, subdir)

            if em_sub_path and accession:
                if em_sub_path not in [
                    "header",
                    "map",
                    "fsc",
                    "images",
                    "masks",
                    "other",
                    "validation",
                ]:
                    raise NameError("em_sub_path %s not allowed" % em_sub_path)

                basedir = os.path.join(basedir, accession.upper(), em_sub_path)

        return basedir

    def get_for_release_path(self):
        """Returns path to for_release directory"""
        return self.__cICommon.get_for_release_path()

    def get_for_release_beta_path(self):
        """Returns path to for_release_beta directory"""
        return self.__cICommon.get_for_release_beta_path()

    def get_for_release_version_path(self):
        """Returns path to for_release_version directory"""
        return self.__cICommon.get_for_release_version_path()

    def get_added_path(self, version=None):
        if version is None:
            version = self.current_folder_name
        return self.getForReleasePath(subdir="added", version=version)

    def get_previous_added_path(self):
        return self.get_added_path(version=self.previous_folder_name)

    def get_modified_path(self, version=None):
        if version is None:
            version = self.current_folder_name
        return self.getForReleasePath(subdir="modified", version=version)

    def get_previous_modified_path(self):
        return self.get_modified_path(version=self.previous_folder_name)

    def get_emd_subfolder_path(self, accession, subfolder, version=None):
        if version is None:
            version = self.current_folder_name
        return self.getForReleasePath(subdir="emd", accession=accession, em_sub_path=subfolder, version=version)

    def get_previous_emd_subfolder_path(self, accession, subfolder):
        return self.get_emd_subfolder_path(accession=accession, subfolder=subfolder, version=self.previous_folder_name)
